@@ -1,5 +1,5 @@
 /*******************************************************************************
-    Copyright (c) 2017 NVIDIA Corporation
+    Copyright (c) 2017-2019 NVIDIA Corporation
 
     Permission is hereby granted, free of charge, to any person obtaining a copy
     of this software and associated documentation files (the "Software"), to
@@ -307,7 +307,7 @@ bool uvm_gpu_access_counters_pending(uvm_gpu_t *gpu)
     if (!gpu->access_counter_buffer_hal->entry_is_valid(gpu, gpu->access_counter_buffer_info.cached_get)) {
         // Slow path: read the put pointer from the GPU register via BAR0 over PCIe
         gpu->access_counter_buffer_info.cached_put =
-            UVM_READ_ONCE(*gpu->access_counter_buffer_info.rm_info.pAccessCntrBufferPut);
+            UVM_GPU_READ_ONCE(*gpu->access_counter_buffer_info.rm_info.pAccessCntrBufferPut);
 
         // No interrupt pending
         if (gpu->access_counter_buffer_info.cached_get == gpu->access_counter_buffer_info.cached_put)
@@ -420,7 +420,7 @@ NV_STATUS uvm_gpu_init_access_counters(uvm_gpu_t *gpu)
     if (granularity_bytes > UVM_MAX_TRANSLATION_SIZE)
         UVM_ASSERT(granularity_bytes % UVM_MAX_TRANSLATION_SIZE == 0);
 
-    gpu->access_counter_buffer_info.ignore_notifications = false;
+    gpu->access_counter_buffer_info.notifications_ignored_count = 0;
     gpu->access_counter_buffer_info.reconfiguration_owner = NULL;
 
     uvm_tracker_init(&access_counters->clear_tracker);
@@ -547,7 +547,7 @@ static NV_STATUS access_counters_take_ownership(uvm_gpu_t *gpu, UvmGpuAccessCntr
     // Read current get pointer as this might not be the first time we have
     // taken control of the notify buffer since the GPU was initialized. Then
     // flush old notifications. This will update the cached_put pointer.
-    access_counters->cached_get = UVM_READ_ONCE(*access_counters->rm_info.pAccessCntrBufferGet);
+    access_counters->cached_get = UVM_GPU_READ_ONCE(*access_counters->rm_info.pAccessCntrBufferGet);
     access_counter_buffer_flush_locked(gpu, UVM_GPU_BUFFER_FLUSH_MODE_UPDATE_PUT);
 
     access_counters->current_config.threshold = config->threshold;
@@ -687,7 +687,7 @@ static void write_get(uvm_gpu_t *gpu, NvU32 get)
     access_counters->cached_get = get;
 
     // Update get pointer on the GPU
-    UVM_WRITE_ONCE(*access_counters->rm_info.pAccessCntrBufferGet, get);
+    UVM_GPU_WRITE_ONCE(*access_counters->rm_info.pAccessCntrBufferGet, get);
 }
 
 static void access_counter_buffer_flush_locked(uvm_gpu_t *gpu, uvm_gpu_buffer_flush_mode_t flush_mode)
@@ -706,7 +706,7 @@ static void access_counter_buffer_flush_locked(uvm_gpu_t *gpu, uvm_gpu_buffer_fl
 
     // Read PUT pointer from the GPU if requested
     if (flush_mode == UVM_GPU_BUFFER_FLUSH_MODE_UPDATE_PUT)
-        access_counters->cached_put = UVM_READ_ONCE(*access_counters->rm_info.pAccessCntrBufferPut);
+        access_counters->cached_put = UVM_GPU_READ_ONCE(*access_counters->rm_info.pAccessCntrBufferPut);
 
     get = access_counters->cached_get;
     put = access_counters->cached_put;
@@ -812,8 +812,7 @@ static NvU32 fetch_access_counter_buffer_entries(uvm_gpu_t *gpu,
 
     // Read put pointer from GPU and cache it
     if (get == access_counters->cached_put) {
-        access_counters->cached_put =
-            UVM_READ_ONCE(*access_counters->rm_info.pAccessCntrBufferPut);
+        access_counters->cached_put = UVM_GPU_READ_ONCE(*access_counters->rm_info.pAccessCntrBufferPut);
     }
 
     put = access_counters->cached_put;
@@ -1469,7 +1468,7 @@ void uvm_gpu_service_access_counters(uvm_gpu_t *gpu)
 
     UVM_ASSERT(gpu->access_counters_supported);
 
-    if (gpu->access_counter_buffer_info.ignore_notifications)
+    if (gpu->access_counter_buffer_info.notifications_ignored_count > 0)
         return;
 
     while (1) {
@@ -1576,11 +1575,12 @@ void uvm_perf_access_counters_unload(uvm_va_space_t *va_space)
 NV_STATUS uvm8_test_access_counters_enabled_by_default(UVM_TEST_ACCESS_COUNTERS_ENABLED_BY_DEFAULT_PARAMS *params,
                                                        struct file *filp)
 {
+    uvm_va_space_t *va_space = uvm_va_space_get(filp);
     uvm_gpu_t *gpu = NULL;
-    NV_STATUS status = uvm_gpu_retain_by_uuid(&params->gpu_uuid, &gpu);
 
-    if (status != NV_OK)
-        return status;
+    gpu = uvm_va_space_retain_gpu_by_uuid(va_space, &params->gpu_uuid);
+    if (!gpu)
+        return NV_ERR_INVALID_DEVICE;
 
     params->enabled = uvm_gpu_access_counters_required(gpu);
 
@@ -1603,9 +1603,9 @@ NV_STATUS uvm8_test_reconfigure_access_counters(UVM_TEST_RECONFIGURE_ACCESS_COUN
     if (status != NV_OK)
         return status;
 
-    status = uvm_gpu_retain_by_uuid(&params->gpu_uuid, &gpu);
-    if (status != NV_OK)
-        return status;
+    gpu = uvm_va_space_retain_gpu_by_uuid(va_space, &params->gpu_uuid);
+    if (!gpu)
+        return NV_ERR_INVALID_DEVICE;
 
     if (!gpu->access_counters_supported) {
         status = NV_ERR_NOT_SUPPORTED;
@@ -1708,6 +1708,7 @@ NV_STATUS uvm8_test_reset_access_counters(UVM_TEST_RESET_ACCESS_COUNTERS_PARAMS
     NV_STATUS status = NV_OK;
     uvm_gpu_t *gpu = NULL;
     uvm_access_counter_buffer_info_t *access_counters;
+    uvm_va_space_t *va_space = uvm_va_space_get(filp);
 
     if (params->mode >= UVM_TEST_ACCESS_COUNTER_RESET_MODE_MAX)
         return NV_ERR_INVALID_ARGUMENT;
@@ -1717,9 +1718,9 @@ NV_STATUS uvm8_test_reset_access_counters(UVM_TEST_RESET_ACCESS_COUNTERS_PARAMS
         return NV_ERR_INVALID_ARGUMENT;
     }
 
-    status = uvm_gpu_retain_by_uuid(&params->gpu_uuid, &gpu);
-    if (status != NV_OK)
-        return status;
+    gpu = uvm_va_space_retain_gpu_by_uuid(va_space, &params->gpu_uuid);
+    if (!gpu)
+        return NV_ERR_INVALID_DEVICE;
 
     if (!gpu->access_counters_supported) {
         status = NV_ERR_NOT_SUPPORTED;
@@ -1765,12 +1766,24 @@ exit_release_gpu:
 
 void uvm_gpu_access_counters_set_ignore(uvm_gpu_t *gpu, bool do_ignore)
 {
+    bool change_intr_state = false;
+
     if (!gpu->access_counters_supported)
         return;
 
     uvm_gpu_access_counters_isr_lock(gpu);
 
-    if (gpu->access_counter_buffer_info.ignore_notifications != do_ignore) {
+    if (do_ignore) {
+        if (gpu->access_counter_buffer_info.notifications_ignored_count++ == 0)
+            change_intr_state = true;
+    }
+    else {
+        UVM_ASSERT(gpu->access_counter_buffer_info.notifications_ignored_count >= 1);
+        if (--gpu->access_counter_buffer_info.notifications_ignored_count == 0)
+            change_intr_state = true;
+    }
+
+    if (change_intr_state) {
         // We need to avoid an interrupt storm while ignoring notifications. We
         // just disable the interrupt.
         uvm_spin_lock_irqsave(&gpu->isr.interrupts_lock);
@@ -1784,8 +1797,6 @@ void uvm_gpu_access_counters_set_ignore(uvm_gpu_t *gpu, bool do_ignore)
 
         if (!do_ignore)
             access_counter_buffer_flush_locked(gpu, UVM_GPU_BUFFER_FLUSH_MODE_CACHED_PUT);
-
-        gpu->access_counter_buffer_info.ignore_notifications = do_ignore;
     }
 
     uvm_gpu_access_counters_isr_unlock(gpu);
@@ -1793,12 +1804,13 @@ void uvm_gpu_access_counters_set_ignore(uvm_gpu_t *gpu, bool do_ignore)
 
 NV_STATUS uvm8_test_set_ignore_access_counters(UVM_TEST_SET_IGNORE_ACCESS_COUNTERS_PARAMS *params, struct file *filp)
 {
+    uvm_va_space_t *va_space = uvm_va_space_get(filp);
     NV_STATUS status = NV_OK;
     uvm_gpu_t *gpu = NULL;
 
-    status = uvm_gpu_retain_by_uuid(&params->gpu_uuid, &gpu);
-    if (status != NV_OK)
-        return status;
+    gpu = uvm_va_space_retain_gpu_by_uuid(va_space, &params->gpu_uuid);
+    if (!gpu)
+        return NV_ERR_INVALID_DEVICE;
 
     if (gpu->access_counters_supported)
         uvm_gpu_access_counters_set_ignore(gpu, params->ignore);
