@@ -1,5 +1,5 @@
 /*******************************************************************************
-    Copyright (c) 2015-2018 NVIDIA Corporation
+    Copyright (c) 2015-2019 NVIDIA Corporation
 
     Permission is hereby granted, free of charge, to any person obtaining a copy
     of this software and associated documentation files (the "Software"), to
@@ -364,10 +364,13 @@ struct uvm_va_space_struct
 
         atomic_t migrate_vma_allocation_fail_nth;
     } test;
+
+    // Queue item for deferred f_ops->release() handling
+    nv_kthread_q_item_t deferred_release_q_item;
 };
 
 NV_STATUS uvm_va_space_create(struct inode *inode, struct file *filp);
-void uvm_va_space_destroy(struct file *filp);
+void uvm_va_space_destroy(uvm_va_space_t *va_space);
 
 // All VA space locking should be done with these wrappers. They're macros so
 // lock assertions are attributed to line numbers correctly.
@@ -430,12 +433,23 @@ void uvm_va_space_destroy(struct file *filp);
 // Get a registered gpu by uuid. This restricts the search for GPUs, to those that
 // have been registered with a va_space. This returns NULL if the GPU is not present, or not
 // registered with the va_space.
+//
+// LOCKING: The VA space lock must be held.
 uvm_gpu_t *uvm_va_space_get_gpu_by_uuid(uvm_va_space_t *va_space, const NvProcessorUuid *gpu_uuid);
 
 // Like uvm_va_space_get_gpu_by_uuid, but also returns NULL if the GPU does
 // not have a GPU VA space registered in the UVM va_space.
+//
+// LOCKING: The VA space lock must be held.
 uvm_gpu_t *uvm_va_space_get_gpu_by_uuid_with_gpu_va_space(uvm_va_space_t *va_space, const NvProcessorUuid *gpu_uuid);
 
+// Same as uvm_va_space_get_gpu_by_uuid but it also retains the GPU. The caller
+// cannot assume that the GPU is still registered in the VA space after the
+// function returns.
+//
+// LOCKING: The function takes and releases the VA space lock in read mode.
+uvm_gpu_t *uvm_va_space_retain_gpu_by_uuid(uvm_va_space_t *va_space, const NvProcessorUuid *gpu_uuid);
+
 // Returns whether read-duplication is supported
 // If gpu is NULL, returns the current state.
 // otherwise, it retuns what the result would be once the gpu's va space is added or removed
@@ -572,11 +586,19 @@ static uvm_gpu_va_space_t *uvm_gpu_va_space_get(uvm_va_space_t *va_space, uvm_gp
 // Return the processor in the candidates mask that is "closest" to src, or
 // UVM_MAX_PROCESSORS if candidates is empty. The order is:
 // - src itself
-// - Direct NVLINK peers (src is CPU or GPU)
-// - Indirect NVLINK peers (src is GPU)
-// - PCIe peers (src is GPU)
-// - CPU (src is GPU)
-// - Arbitrary selection
+// - Direct NVLINK GPU peers if src is CPU or GPU (1)
+// - NVLINK CPU if src is GPU
+// - Indirect NVLINK GPU peers if src is GPU
+// - PCIe peers if src is GPU (2)
+// - CPU if src is GPU
+// - Deterministic selection from the pool of candidates
+//
+// (1) When src is a GPU, NVLINK GPU peers are preferred over the CPU because in
+//     NUMA systems the CPU processor may refer to multiple CPU NUMA nodes, and
+//     the bandwidth between src and the farthest CPU node can be substantially
+//     lower than the bandwidth src and its peer GPUs.
+// (2) TODO: Bug 1764943: Is copying from a PCI peer always better than copying
+//     from CPU?
 uvm_processor_id_t uvm_processor_mask_find_closest_id(uvm_va_space_t *va_space,
                                                       const uvm_processor_mask_t *candidates,
                                                       uvm_processor_id_t src);
